import os
import re
from datetime import datetime, timedelta


class EnvExpanderError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return repr(self.message)

class EnvExpander(object):
    def __init__(self, environ=os.environ):
        self.environ = environ
        self.pattern_env = re.compile(r'\$\{(.*?)\}')
        # TODO(Yancey1989): support more date expression
        self.pattern_date_expr = re.compile(r"(yyyyMMdd)\Z|(yyyyMMdd)\s*(\+|\-)\s*(\d+)d")

        
    def _match_date_expr(self, expr):
        return re.match(self.pattern_date_expr, expr)

    def parse_bizdate(self, expr):
        bizdate = self.environ["BIZDATE"][:8]
        if not bizdate:
            raise EnvExpanderError("The date format ${yyyyMMdd +/- 1d} need set the OS environment variable ${BIZDATE}")
        dt = datetime.strptime(bizdate, "%Y%m%d")
        m = self._match_date_expr(expr)
        if m.groups()[0]:
            return dt.strftime("%Y%m%d") 
        else:
            if m.groups()[2] == "+":
                dt += timedelta(days=int(m.groups()[3]))
            elif m.groups()[2] == "-":
                dt -= timedelta(days=int(m.groups()[3]))
            else:
                raise EnvExpanderError("date format failed: {}".format(expr))
            return dt.strftime("%Y%m%d")

    def expand(self, sql):
        new_sql = sql
        for m in re.finditer(self.pattern_env, sql):
            if m.group(1) in self.environ:
                val = self.environ[m.group(1)]
                new_sql = new_sql.replace(m.group(0), val, 1)
            elif self._match_date_expr(m.group(1)):
                val = self.parse_bizdate(m.group(1))
                new_sql = new_sql.replace(m.group(0), val, 1)
            else:
                raise EnvExpanderError("Can not find the environment: {} in the runtime envionrment.".format(m.group(1)))
        return new_sql


class Rows:
    def __init__(self, column_names, rows_gen):
        """Query result of sqlflow.client.Client.execute

        :param column_names: column names
        :type column_names: list[str].
        :param rows_gen: rows generator
        :type rows_gen: generator
        """
        self._column_names = column_names
        self._rows_gen = rows_gen
        self._rows = None

    def column_names(self):
        """Column names

        :return: list[str]
        """
        return self._column_names

    def rows(self):
        """Rows

        Example:

        >>> [r for r in rows.rows()]

        :return: list generator
        """
        if self._rows is None:
            self._rows = []
            for row in self._rows_gen():
                self._rows.append(row)
                yield row
        else:
            for row in self._rows:
                yield row

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return self.to_dataframe().__repr__()

    def _repr_html_(self):
        return self.to_dataframe()._repr_html_()

    def to_dataframe(self):
        """Convert Rows to pandas.Dataframe

        :return: pandas.Dataframe
        """
        # Fetch all the rows to self._rows
        for r in self.rows():
            pass
        import pandas as pd
        return pd.DataFrame(self._rows, columns=self._column_names)




import os
import sys
import logging
import grpc
import re
import time

import google.protobuf.wrappers_pb2 as wrapper
from google.protobuf.timestamp_pb2 import Timestamp

import sqlflow.proto.sqlflow_pb2 as pb
import sqlflow.proto.sqlflow_pb2_grpc as pb_grpc

from sqlflow.env_expand import EnvExpander, EnvExpanderError
from sqlflow.rows import Rows
from sqlflow.compound_message import CompoundMessage

_LOGGER = logging.getLogger(__name__)
handler = logging.StreamHandler(sys.stdout)
_LOGGER.setLevel(logging.INFO)
_LOGGER.addHandler(handler)
# default timeout is 10 hours to tolerate waiting training
# jobs to finish.
DEFAULT_TIMEOUT=3600 * 10


class StreamReader(object):
    def __init__(self, stream_response):
        self._stream_response = stream_response
        self.last_response = None

    def read_one(self):
        try:
            res = next(self._stream_response)
        except StopIteration:
            return (None, None)

        return (res, res.WhichOneof('response'))

    def read_until_type_changed(self):
        first_rtype = None
        while True:
            try:
                response = next(self._stream_response)
            except StopIteration:
                break

            rtype = response.WhichOneof('response')
            if first_rtype == None:
                first_rtype = rtype 

            if first_rtype != rtype:
                self.last_response = response
                break 
            yield response

class Client:
    def __init__(self, server_url=None, ca_crt=None):
        """A minimum client that issues queries to and fetch results/logs from sqlflowserver.

        :param server_url: sqlflowserver url. If None, read value from
                           environment variable SQLFLOW_SERVER.
        :type server_url: str.

        :param ca_crt: Path to CA certificates of SQLFlow client, if None,
                       try to find the file from the environment variable:
                       SQLFLOW_CA_CRT, otherwise using insecure client.
        :type ca_crt: str.

        :raises: ValueError

        Example:
        >>> client = sqlflow.Client(server_url="localhost:50051")

        """
        if server_url is None:
            if "SQLFLOW_SERVER" not in os.environ:
                raise ValueError("Can't find environment variable SQLFLOW_SERVER")
            server_url = os.environ["SQLFLOW_SERVER"]

        self._stub = pb_grpc.SQLFlowStub(self.new_rpc_channel(server_url, ca_crt))
        self._expander = EnvExpander(os.environ)

    def new_rpc_channel(self, server_url, ca_crt):
        if ca_crt is None and "SQLFLOW_CA_CRT" not in os.environ:
            # client would connect SQLFLow gRPC server with insecure mode.
            channel = grpc.insecure_channel(server_url)
        else:
            if ca_crt is None:
                ca_crt = os.environ["SQLFLOW_CA_CRT"]
            with open(ca_crt, "rb") as f:
                creds = grpc.ssl_channel_credentials(f.read())
            channel = grpc.secure_channel(server_url, creds)
        return channel

    def sql_request(self, sql):
        token = os.getenv("SQLFLOW_USER_TOKEN", "")
        db_conn_str = os.getenv("SQLFLOW_DATASOURCE", "")
        exit_on_submit_env = os.getenv("SQLFLOW_EXIT_ON_SUBMIT", "True")
        user_id = os.getenv("SQLFLOW_USER_ID", "")
        hive_location = os.getenv("SQLFLOW_HIVE_LOCATION", "")
        hdfs_namenode_addr = os.getenv("SQLFLOW_HDFS_NAMENODE_ADDR", "")
        # environment variables JUPYTER_HADOOP_USER, JUPYTER_HADOOP_PASS stores the user's hdfs credentials.
        hdfs_user = os.getenv("JUPYTER_HADOOP_USER", "")
        hdfs_pass = os.getenv("JUPYTER_HADOOP_PASS", "")
        if exit_on_submit_env.isdigit():
            exit_on_submit = bool(int(exit_on_submit_env))
        else:
            exit_on_submit = exit_on_submit_env.lower() == "true"
        se = pb.Session(token=token,
                        db_conn_str=db_conn_str,
                        exit_on_submit=exit_on_submit,
                        user_id=user_id,
                        hive_location=hive_location,
                        hdfs_namenode_addr=hdfs_namenode_addr,
                        hdfs_user=hdfs_user,
                        hdfs_pass=hdfs_pass)
        try:
            sql = self._expander.expand(sql)
        except Exception as e:
            _LOGGER.error("%s", e)
            raise e
        return pb.Request(sql=sql, session=se)

    def execute(self, operation):
        """Run a SQL statement

        :param operation: SQL statement to be executed.
        :type operation: str.

        :returns: sqlflow.client.Rows

        Example:

        >>> client.execute("select * from iris limit 1")

        """
        try:
            stream_response = self._stub.Run(self.sql_request(operation), timeout=DEFAULT_TIMEOUT)
            return self.display(stream_response)
        except grpc.RpcError as e:
            # NOTE: raise exception to interrupt notebook execution. Or
            # the notebook will continue execute the next block.
            raise e
        except EnvExpanderError as e:
            raise e

    def read_fetch_response(self, job_id):
        req = pb.FetchRequest()
        pb_job = pb.Job(id=job_id)
        pb_job.id = job_id
        req.job.CopyFrom(pb_job)
        compound_message = CompoundMessage()
        column_names = None
        rows = []
        
        # TODO(yancey1989): using the common codebase with the stream response
        while True:
            fetch_response = self._stub.Fetch(req)
            for response in fetch_response.responses.response:
                rtype = response.WhichOneof('response')
                if rtype == 'message':
                    msg = response.message.message
                    if re.match(r'<[a-z][\s\S]*>.*', msg):
                        from IPython.core.display import display, HTML
                        display(HTML(msg))
                    else:
                        _LOGGER.info(msg)
                elif rtype == 'eoe':
                    def _rows_gen():
                        for row in rows:
                            yield [self._decode_any(d) for d in row.data]
                    compound_message.add_rows(Rows(column_names, _rows_gen), None)
                    rows = []
                    break
                elif rtype == 'head':
                    column_names = [column_name for column_name in response.head.column_names]
                else:
                    rows.append(response.row)
            if fetch_response.eof:
                break
            req = fetch_response.updated_fetch_since
            time.sleep(2)
        
        return compound_message

    def display_html(self, first_line, stream_reader):
        resp_list = [first_line]
        for res in stream_reader.read_until_type_changed():
            resp_list.append(res.message.message)
        from IPython.core.display import display, HTML
        display(HTML('\n'.join(resp_list)))

    def display(self, stream_response):
        """Display stream response like log or table.row"""

        reader = StreamReader(stream_response)
        response, rtype = reader.read_one()
        compound_message = CompoundMessage()
        while True:
            if response == None:
                break
            if rtype == 'message':
                if re.match(r'<[a-z][\s\S]*>.*', response.message.message):
                    self.display_html(response.message.message, reader)
                else:
                    _LOGGER.info(response.message.message)
                    for response in reader.read_until_type_changed():
                        _LOGGER.info(response.message.message)
                    response = reader.last_response
                    if response is not None:
                        rtype = response.WhichOneof('response')
                    continue
            elif rtype == 'job':
                job = response.job
                # the last response type is Job for the workflow mode,
                # so break the loop here
                return self.read_fetch_response(job.id)
            else:
                column_names = [column_name for column_name in response.head.column_names]
                def rows_gen():
                    for res in reader.read_until_type_changed():
                        yield [self._decode_any(a) for a in res.row.data]
                compound_message.add_rows(Rows(column_names, rows_gen), None)

            # read the next response
            response, rtype = reader.read_one()

        return compound_message
    
    @classmethod
    def _decode_any(cls, any_message):
        """Decode a google.protobuf.any_pb2
        """
        try:
            message = next(getattr(wrapper, type_name)()
                           for type_name, desc in wrapper.DESCRIPTOR.message_types_by_name.items()
                           if any_message.Is(desc))
            any_message.Unpack(message)
            return message.value
        except StopIteration:
            if any_message.Is(pb.Row.Null.DESCRIPTOR):
                return None
            if any_message.Is(Timestamp.DESCRIPTOR):
                timestamp_message = Timestamp()
                any_message.Unpack(timestamp_message)
                return timestamp_message.ToDatetime()
            raise TypeError("Unsupported type {}".format(any_message))



from sqlflow.rows import Rows

class CompoundMessage:
    def __init__(self):
        """Message containing return result of several SQL statements
           CompoundMessage can not display in notebook since we need to
           output log messages for long running training sqls.
        """
        self._messages = []
        self.TypeRows = 1
        self.TypeMessage = 2
        self.TypeHTML = 3
    
    def add_rows(self, rows, eoe):
        assert(isinstance(rows, Rows))
        # call __str__() to trigger rows_gen
        rows.__str__()
        self._messages.append((rows, eoe, self.TypeRows))
    
    def add_message(self, message, eoe):
        assert(isinstance(message, str))
        self._messages.append((message, eoe, self.TypeMessage))
    
    def add_html(self, message, eoe):
        assert(isinstance(message, str))
        self._messages.append((message, eoe, self.TypeHTML))
    
    def length(self):
        return len(self._messages)

    def __str__(self):
        return self.__repr__()
    
    def __repr__(self):
        all_string = ""
        for r in self._messages:
            if isinstance(r[0], Rows):
                all_string = '\n'.join([all_string, r[0].__repr__()])
            else:
                all_string = '\n'.join([all_string, r[0].__repr__()])
        return all_string

    def _repr_html_(self):
        all_html = ""
        for r in self._messages:
            if isinstance(r[0], Rows):
                all_html = ''.join([all_html, r[0]._repr_html_()])
            else:
                all_html = ''.join([all_html, "<p>%s</p>" % (r[0].__str__().replace("\n", "<br>"))])
        return all_html

    def get(self, idx):
        return self._messages[idx][0]




# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from sqlflow.proto import sqlflow_pb2 as sqlflow_dot_proto_dot_sqlflow__pb2


class SQLFlowStub(object):
    """Missing associated documentation comment in .proto file"""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Run = channel.unary_stream(
                '/proto.SQLFlow/Run',
                request_serializer=sqlflow_dot_proto_dot_sqlflow__pb2.Request.SerializeToString,
                response_deserializer=sqlflow_dot_proto_dot_sqlflow__pb2.Response.FromString,
                )
        self.Fetch = channel.unary_unary(
                '/proto.SQLFlow/Fetch',
                request_serializer=sqlflow_dot_proto_dot_sqlflow__pb2.FetchRequest.SerializeToString,
                response_deserializer=sqlflow_dot_proto_dot_sqlflow__pb2.FetchResponse.FromString,
                )


class SQLFlowServicer(object):
    """Missing associated documentation comment in .proto file"""

    def Run(self, request, context):
        """Run executes a sql statement

        SQL statements like `SELECT ...`, `DESCRIBE ...` returns a rowset.
        The rowset might be big. In such cases, Query returns a stream
        of RunResponse

        SQLFlow implements the Run interface with two mode:

        1. Local model
        The SQLFlow server execute the SQL statements on the local host.

        SQL statements like `USE database`, `DELETE` returns only a success
        message.

        SQL statement like `SELECT ... TO TRAIN/PREDICT ...` returns a stream of
        messages which indicates the training/predicting progress

        2. Argo Workflow mode
        The SQLFlow server submits an Argo workflow into a Kubernetes cluster,
        and returns a stream of messages indicates the WorkFlow ID and the 
        submitting progress.

        The SQLFlow gRPC client should fetch the logs of the workflow by
        calling the Fetch interface in a polling manner.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Fetch(self, request, context):
        """Fetch fetches the SQLFlow job phase and logs in a polling manner. A corresponding
        client can be implemented as

        wfJob := Submit(argoYAML)
        req := &FetchRequest {
        Job : { Id: wfJob },
        }
        for {
        res := Fetch(req)
        fmt.Println(res.Logs)
        if isComplete(res) {
        break
        }
        req = res.UpdatedFetchSince
        time.Sleep(time.Second)
        }

        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_SQLFlowServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Run': grpc.unary_stream_rpc_method_handler(
                    servicer.Run,
                    request_deserializer=sqlflow_dot_proto_dot_sqlflow__pb2.Request.FromString,
                    response_serializer=sqlflow_dot_proto_dot_sqlflow__pb2.Response.SerializeToString,
            ),
            'Fetch': grpc.unary_unary_rpc_method_handler(
                    servicer.Fetch,
                    request_deserializer=sqlflow_dot_proto_dot_sqlflow__pb2.FetchRequest.FromString,
                    response_serializer=sqlflow_dot_proto_dot_sqlflow__pb2.FetchResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'proto.SQLFlow', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class SQLFlow(object):
    """Missing associated documentation comment in .proto file"""

    @staticmethod
    def Run(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/proto.SQLFlow/Run',
            sqlflow_dot_proto_dot_sqlflow__pb2.Request.SerializeToString,
            sqlflow_dot_proto_dot_sqlflow__pb2.Response.FromString,
            options, channel_credentials,
            call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Fetch(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/proto.SQLFlow/Fetch',
            sqlflow_dot_proto_dot_sqlflow__pb2.FetchRequest.SerializeToString,
            sqlflow_dot_proto_dot_sqlflow__pb2.FetchResponse.FromString,
            options, channel_credentials,
            call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sqlflow/proto/sqlflow.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='sqlflow/proto/sqlflow.proto',
  package='proto',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=b'\n\x1bsqlflow/proto/sqlflow.proto\x12\x05proto\x1a\x19google/protobuf/any.proto\"\x11\n\x03Job\x12\n\n\x02id\x18\x01 \x01(\t\"L\n\x0c\x46\x65tchRequest\x12\x17\n\x03job\x18\x01 \x01(\x0b\x32\n.proto.Job\x12\x0f\n\x07step_id\x18\x02 \x01(\t\x12\x12\n\nstep_phase\x18\x03 \x01(\t\"\xb1\x01\n\rFetchResponse\x12\x30\n\x13updated_fetch_since\x18\x01 \x01(\x0b\x32\x13.proto.FetchRequest\x12\x0b\n\x03\x65of\x18\x02 \x01(\x08\x12\x31\n\tresponses\x18\x04 \x01(\x0b\x32\x1e.proto.FetchResponse.Responses\x1a.\n\tResponses\x12!\n\x08response\x18\x01 \x03(\x0b\x32\x0f.proto.Response\"\xc2\x01\n\x07Session\x12\r\n\x05token\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x62_conn_str\x18\x02 \x01(\t\x12\x16\n\x0e\x65xit_on_submit\x18\x03 \x01(\x08\x12\x0f\n\x07user_id\x18\x04 \x01(\t\x12\x15\n\rhive_location\x18\x05 \x01(\t\x12\x1a\n\x12hdfs_namenode_addr\x18\x06 \x01(\t\x12\x11\n\thdfs_user\x18\x07 \x01(\t\x12\x11\n\thdfs_pass\x18\x08 \x01(\t\x12\x11\n\tsubmitter\x18\t \x01(\t\"7\n\x07Request\x12\x0b\n\x03sql\x18\x01 \x01(\t\x12\x1f\n\x07session\x18\x02 \x01(\x0b\x32\x0e.proto.Session\"\xb2\x01\n\x08Response\x12\x1b\n\x04head\x18\x01 \x01(\x0b\x32\x0b.proto.HeadH\x00\x12\x19\n\x03row\x18\x02 \x01(\x0b\x32\n.proto.RowH\x00\x12!\n\x07message\x18\x03 \x01(\x0b\x32\x0e.proto.MessageH\x00\x12$\n\x03\x65oe\x18\x04 \x01(\x0b\x32\x15.proto.EndOfExecutionH\x00\x12\x19\n\x03job\x18\x05 \x01(\x0b\x32\n.proto.JobH\x00\x42\n\n\x08response\"\x1c\n\x04Head\x12\x14\n\x0c\x63olumn_names\x18\x01 \x03(\t\"1\n\x03Row\x12\"\n\x04\x64\x61ta\x18\x01 \x03(\x0b\x32\x14.google.protobuf.Any\x1a\x06\n\x04Null\"\x1a\n\x07Message\x12\x0f\n\x07message\x18\x01 \x01(\t\"9\n\x0e\x45ndOfExecution\x12\x0b\n\x03sql\x18\x01 \x01(\t\x12\x1a\n\x12spent_time_seconds\x18\x02 \x01(\x03\x32g\n\x07SQLFlow\x12(\n\x03Run\x12\x0e.proto.Request\x1a\x0f.proto.Response0\x01\x12\x32\n\x05\x46\x65tch\x12\x13.proto.FetchRequest\x1a\x14.proto.FetchResponseb\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_any__pb2.DESCRIPTOR,])




_JOB = _descriptor.Descriptor(
  name='Job',
  full_name='proto.Job',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='proto.Job.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=65,
  serialized_end=82,
)


_FETCHREQUEST = _descriptor.Descriptor(
  name='FetchRequest',
  full_name='proto.FetchRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='job', full_name='proto.FetchRequest.job', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='step_id', full_name='proto.FetchRequest.step_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='step_phase', full_name='proto.FetchRequest.step_phase', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=84,
  serialized_end=160,
)


_FETCHRESPONSE_RESPONSES = _descriptor.Descriptor(
  name='Responses',
  full_name='proto.FetchResponse.Responses',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response', full_name='proto.FetchResponse.Responses.response', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=294,
  serialized_end=340,
)

_FETCHRESPONSE = _descriptor.Descriptor(
  name='FetchResponse',
  full_name='proto.FetchResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='updated_fetch_since', full_name='proto.FetchResponse.updated_fetch_since', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='eof', full_name='proto.FetchResponse.eof', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='responses', full_name='proto.FetchResponse.responses', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_FETCHRESPONSE_RESPONSES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=163,
  serialized_end=340,
)


_SESSION = _descriptor.Descriptor(
  name='Session',
  full_name='proto.Session',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='token', full_name='proto.Session.token', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='db_conn_str', full_name='proto.Session.db_conn_str', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='exit_on_submit', full_name='proto.Session.exit_on_submit', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='user_id', full_name='proto.Session.user_id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hive_location', full_name='proto.Session.hive_location', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hdfs_namenode_addr', full_name='proto.Session.hdfs_namenode_addr', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hdfs_user', full_name='proto.Session.hdfs_user', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hdfs_pass', full_name='proto.Session.hdfs_pass', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='submitter', full_name='proto.Session.submitter', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=343,
  serialized_end=537,
)


_REQUEST = _descriptor.Descriptor(
  name='Request',
  full_name='proto.Request',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='sql', full_name='proto.Request.sql', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='session', full_name='proto.Request.session', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=539,
  serialized_end=594,
)


_RESPONSE = _descriptor.Descriptor(
  name='Response',
  full_name='proto.Response',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='head', full_name='proto.Response.head', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='row', full_name='proto.Response.row', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='proto.Response.message', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='eoe', full_name='proto.Response.eoe', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='job', full_name='proto.Response.job', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='response', full_name='proto.Response.response',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=597,
  serialized_end=775,
)


_HEAD = _descriptor.Descriptor(
  name='Head',
  full_name='proto.Head',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='column_names', full_name='proto.Head.column_names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=777,
  serialized_end=805,
)


_ROW_NULL = _descriptor.Descriptor(
  name='Null',
  full_name='proto.Row.Null',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=850,
  serialized_end=856,
)

_ROW = _descriptor.Descriptor(
  name='Row',
  full_name='proto.Row',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='proto.Row.data', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_ROW_NULL, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=807,
  serialized_end=856,
)


_MESSAGE = _descriptor.Descriptor(
  name='Message',
  full_name='proto.Message',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='proto.Message.message', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=858,
  serialized_end=884,
)


_ENDOFEXECUTION = _descriptor.Descriptor(
  name='EndOfExecution',
  full_name='proto.EndOfExecution',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='sql', full_name='proto.EndOfExecution.sql', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='spent_time_seconds', full_name='proto.EndOfExecution.spent_time_seconds', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=886,
  serialized_end=943,
)

_FETCHREQUEST.fields_by_name['job'].message_type = _JOB
_FETCHRESPONSE_RESPONSES.fields_by_name['response'].message_type = _RESPONSE
_FETCHRESPONSE_RESPONSES.containing_type = _FETCHRESPONSE
_FETCHRESPONSE.fields_by_name['updated_fetch_since'].message_type = _FETCHREQUEST
_FETCHRESPONSE.fields_by_name['responses'].message_type = _FETCHRESPONSE_RESPONSES
_REQUEST.fields_by_name['session'].message_type = _SESSION
_RESPONSE.fields_by_name['head'].message_type = _HEAD
_RESPONSE.fields_by_name['row'].message_type = _ROW
_RESPONSE.fields_by_name['message'].message_type = _MESSAGE
_RESPONSE.fields_by_name['eoe'].message_type = _ENDOFEXECUTION
_RESPONSE.fields_by_name['job'].message_type = _JOB
_RESPONSE.oneofs_by_name['response'].fields.append(
  _RESPONSE.fields_by_name['head'])
_RESPONSE.fields_by_name['head'].containing_oneof = _RESPONSE.oneofs_by_name['response']
_RESPONSE.oneofs_by_name['response'].fields.append(
  _RESPONSE.fields_by_name['row'])
_RESPONSE.fields_by_name['row'].containing_oneof = _RESPONSE.oneofs_by_name['response']
_RESPONSE.oneofs_by_name['response'].fields.append(
  _RESPONSE.fields_by_name['message'])
_RESPONSE.fields_by_name['message'].containing_oneof = _RESPONSE.oneofs_by_name['response']
_RESPONSE.oneofs_by_name['response'].fields.append(
  _RESPONSE.fields_by_name['eoe'])
_RESPONSE.fields_by_name['eoe'].containing_oneof = _RESPONSE.oneofs_by_name['response']
_RESPONSE.oneofs_by_name['response'].fields.append(
  _RESPONSE.fields_by_name['job'])
_RESPONSE.fields_by_name['job'].containing_oneof = _RESPONSE.oneofs_by_name['response']
_ROW_NULL.containing_type = _ROW
_ROW.fields_by_name['data'].message_type = google_dot_protobuf_dot_any__pb2._ANY
DESCRIPTOR.message_types_by_name['Job'] = _JOB
DESCRIPTOR.message_types_by_name['FetchRequest'] = _FETCHREQUEST
DESCRIPTOR.message_types_by_name['FetchResponse'] = _FETCHRESPONSE
DESCRIPTOR.message_types_by_name['Session'] = _SESSION
DESCRIPTOR.message_types_by_name['Request'] = _REQUEST
DESCRIPTOR.message_types_by_name['Response'] = _RESPONSE
DESCRIPTOR.message_types_by_name['Head'] = _HEAD
DESCRIPTOR.message_types_by_name['Row'] = _ROW
DESCRIPTOR.message_types_by_name['Message'] = _MESSAGE
DESCRIPTOR.message_types_by_name['EndOfExecution'] = _ENDOFEXECUTION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Job = _reflection.GeneratedProtocolMessageType('Job', (_message.Message,), {
  'DESCRIPTOR' : _JOB,
  '__module__' : 'sqlflow.proto.sqlflow_pb2'
  # @@protoc_insertion_point(class_scope:proto.Job)
  })
_sym_db.RegisterMessage(Job)

FetchRequest = _reflection.GeneratedProtocolMessageType('FetchRequest', (_message.Message,), {
  'DESCRIPTOR' : _FETCHREQUEST,
  '__module__' : 'sqlflow.proto.sqlflow_pb2'
  # @@protoc_insertion_point(class_scope:proto.FetchRequest)
  })
_sym_db.RegisterMessage(FetchRequest)

FetchResponse = _reflection.GeneratedProtocolMessageType('FetchResponse', (_message.Message,), {

  'Responses' : _reflection.GeneratedProtocolMessageType('Responses', (_message.Message,), {
    'DESCRIPTOR' : _FETCHRESPONSE_RESPONSES,
    '__module__' : 'sqlflow.proto.sqlflow_pb2'
    # @@protoc_insertion_point(class_scope:proto.FetchResponse.Responses)
    })
  ,
  'DESCRIPTOR' : _FETCHRESPONSE,
  '__module__' : 'sqlflow.proto.sqlflow_pb2'
  # @@protoc_insertion_point(class_scope:proto.FetchResponse)
  })
_sym_db.RegisterMessage(FetchResponse)
_sym_db.RegisterMessage(FetchResponse.Responses)

Session = _reflection.GeneratedProtocolMessageType('Session', (_message.Message,), {
  'DESCRIPTOR' : _SESSION,
  '__module__' : 'sqlflow.proto.sqlflow_pb2'
  # @@protoc_insertion_point(class_scope:proto.Session)
  })
_sym_db.RegisterMessage(Session)

Request = _reflection.GeneratedProtocolMessageType('Request', (_message.Message,), {
  'DESCRIPTOR' : _REQUEST,
  '__module__' : 'sqlflow.proto.sqlflow_pb2'
  # @@protoc_insertion_point(class_scope:proto.Request)
  })
_sym_db.RegisterMessage(Request)

Response = _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
  'DESCRIPTOR' : _RESPONSE,
  '__module__' : 'sqlflow.proto.sqlflow_pb2'
  # @@protoc_insertion_point(class_scope:proto.Response)
  })
_sym_db.RegisterMessage(Response)

Head = _reflection.GeneratedProtocolMessageType('Head', (_message.Message,), {
  'DESCRIPTOR' : _HEAD,
  '__module__' : 'sqlflow.proto.sqlflow_pb2'
  # @@protoc_insertion_point(class_scope:proto.Head)
  })
_sym_db.RegisterMessage(Head)

Row = _reflection.GeneratedProtocolMessageType('Row', (_message.Message,), {

  'Null' : _reflection.GeneratedProtocolMessageType('Null', (_message.Message,), {
    'DESCRIPTOR' : _ROW_NULL,
    '__module__' : 'sqlflow.proto.sqlflow_pb2'
    # @@protoc_insertion_point(class_scope:proto.Row.Null)
    })
  ,
  'DESCRIPTOR' : _ROW,
  '__module__' : 'sqlflow.proto.sqlflow_pb2'
  # @@protoc_insertion_point(class_scope:proto.Row)
  })
_sym_db.RegisterMessage(Row)
_sym_db.RegisterMessage(Row.Null)

Message = _reflection.GeneratedProtocolMessageType('Message', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGE,
  '__module__' : 'sqlflow.proto.sqlflow_pb2'
  # @@protoc_insertion_point(class_scope:proto.Message)
  })
_sym_db.RegisterMessage(Message)

EndOfExecution = _reflection.GeneratedProtocolMessageType('EndOfExecution', (_message.Message,), {
  'DESCRIPTOR' : _ENDOFEXECUTION,
  '__module__' : 'sqlflow.proto.sqlflow_pb2'
  # @@protoc_insertion_point(class_scope:proto.EndOfExecution)
  })
_sym_db.RegisterMessage(EndOfExecution)



_SQLFLOW = _descriptor.ServiceDescriptor(
  name='SQLFlow',
  full_name='proto.SQLFlow',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=945,
  serialized_end=1048,
  methods=[
  _descriptor.MethodDescriptor(
    name='Run',
    full_name='proto.SQLFlow.Run',
    index=0,
    containing_service=None,
    input_type=_REQUEST,
    output_type=_RESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='Fetch',
    full_name='proto.SQLFlow.Fetch',
    index=1,
    containing_service=None,
    input_type=_FETCHREQUEST,
    output_type=_FETCHRESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_SQLFLOW)

DESCRIPTOR.services_by_name['SQLFlow'] = _SQLFLOW

# @@protoc_insertion_point(module_scope)



import argparse

from sqlflow.client import Client

parser = argparse.ArgumentParser()
parser.add_argument('sql', nargs='+', type=str, help="sql", action="store")
parser.add_argument("--url", type=str, help="server url", action="store", default=None)
parser.add_argument("--ca_crt", type=str, help="Path to CA certificates of SQLFlow client.", action="store", default=None)

def main():
    args = parser.parse_args()

    client = Client(server_url=args.url, ca_crt=args.ca_crt)
    for sql in args.sql:
        print("executing: {}".format(sql))
        print(client.execute(sql))



import sys
import logging

from IPython.core.magic import Magics, magics_class, cell_magic, line_magic
from sqlflow.client import Client, _LOGGER

# For changing input cell syntax highlighting logic for the entire session
# http://stackoverflow.com/questions/28703626/ipython-change-input-cell-syntax-highlighting-logic-for-entire-session
from IPython.display import display_javascript

@magics_class
class SqlFlowMagic(Magics):
    """
    Provides the `%%sqlflow` magic
    """
    def __init__(self, shell):
        super(SqlFlowMagic, self).__init__(shell)
        self.client = Client()

    @cell_magic('sqlflow')
    def execute(self, line, cell):
        """Runs SQL statement

        :param line: The line magic
        :type line: str.
        :param cell: The cell magic
        :type cell: str.

        Example:

        >>> %%sqlflow SELECT *
        ... FROM mytable

        >>> %%sqlflow SELECT *
        ... FROM iris.iris limit 1
        ... TRAIN DNNClassifier
        ... WITH
        ...   n_classes = 3,
        ...   hidden_units = [10, 10]
        ... COLUMN sepal_length, sepal_width, petal_length, petal_width
        ... LABEL class
        ... INTO my_dnn_model;

        """
        return self.client.execute('\n'.join([line, cell]))

def load_ipython_extension(ipython):
    # Change input cell syntax highlighting to SQL
    js = "IPython.CodeCell.options_default.highlight_modes['magic_sql'] = {'reg':[/^%%sqlflow/]};"
    display_javascript(js, raw=True)

    magics = SqlFlowMagic(ipython)
    ipython.register_magics(magics)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os

from setuptools import find_packages, setup

# Package meta-data.
NAME = 'sqlflow'
DESCRIPTION = 'SQLFlow client library for Python.'
URL = 'https://github.com/sql-machine-learning/sqlflow'
EMAIL = 'kuisong.tong@gmail.com'
AUTHOR = 'Kuisong Tong'
REQUIRES_PYTHON = '>=3.5.0'
VERSION = None

# What packages are required for this module to be executed?
REQUIRED = [
    'protobuf >=3.6, <4',
    'grpcio >=1.17, <2',
    'ipython==7.9',
    'pandas',
]
SETUP_REQUIRED = [
    'pytest-runner'
]
TEST_REQUIRED = [
    'pytest',
]

# What packages are optional?
EXTRAS = {
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's _version.py module as a dictionary.
about = {}
if not VERSION:
    with open(os.path.join(here, NAME, '_version.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION

# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    package_data={'sqlflow': ['proto/*.py']},
    entry_points={
        'console_scripts': ['sqlflow = sqlflow.__main__:main'],
    },
    install_requires=REQUIRED,
    setup_requires=SETUP_REQUIRED,
    tests_require=TEST_REQUIRED,
    extras_require=EXTRAS,
    license='Apache License 2.0',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
)
